/* fehler.c
 * Eingabe:  zwei Wertetabellen
 * Ausgabe:  maximale Abweichung in den Y-Werten
 * erstellt: Die Nov 23 18:23:05 CET 1999
 * Autoren : Mirko Belickin und Joachim Wagner
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "matrix.h"
#include "nmmatrix.h"

extern matrix allg_matrix;

/* Hauptprogramm 
 */

int main(int argc, char *argv[])
{
  FILE *ist, *soll;
  int hilfe = 0;        /* 1 = Hilfe fuer Benutzer anzeigen */   
  int fehler = 0;       /* 1 = Programm falsch aufgerufen */
  int schalter = 1;     /* 0 = es folgen keine Schalter ('--' gefunden) */
  int i, n;
  double max, q;        /* Hilfsvariablen Maximumsbestimmung */
  matrix *y1, *y2;
  
  while ((schalter) && (argc>1) && (*argv[1]=='-')) {/* Schalter bearbeiten */
    switch ((argv[1])[1]) {
      case 'h': hilfe = 1;
                break;
      case '-': schalter = 0;
                break;
      default:  fprintf(stderr,"Schalter %s unbekannt.\n", argv[1]);
                fehler = 1;
                break;
    }
    argv[1] = argv[0];
    argv++; 
    argc--;
  }
  
  if ((argc>3) || (argc<3)) fehler = 1;
  if ((hilfe) || (fehler)) {             /* Hilfetext anzeigen */
    fprintf(stderr,"\nBenutzung: "
                   "%s [Schalter] [--] Sollwertedatei Istwertedatei\n\n",
                   argv[0]);
    fprintf(stderr,"Schalter: -h fuer diese Hilfe\n");
    fprintf(stderr,"\nMirko Belickin und Joachim Wagner - 23. November 1999\n");
    exit(fehler);
  }
  if (!(soll=fopen(argv[1],"rb"))) {
    fprintf(stderr,"Oeffnen der Datei %s fehlgeschlagen.", argv[1]);
    exit(1);
  }
  if (!(ist=fopen(argv[2],"rb"))) {
    fprintf(stderr,"Oeffnen der Datei %s fehlgeschlagen.", argv[2]);
    exit(1);
  }
  
  y1 = allg_matrix.aktion->laden(soll, NULL);  /* X-Werte lesen */
  y1->aktion->freigeben(y1, NULL);             /* und verwerfen */
  y1 = allg_matrix.aktion->laden(soll, NULL);  /* Y-Werte */
  y2 = allg_matrix.aktion->laden(ist, NULL);   /* X-Werte lesen */
  y2->aktion->freigeben(y2, NULL);             /* und verwerfen */
  y2 = allg_matrix.aktion->laden(ist, NULL);   /* Y-Werte */
  fclose(soll);
  fclose(ist);

  n = y1->aktion->hoehe(y1);
  if (n != y2->aktion->hoehe(y2)) {
    fprintf(stderr,"Anzahl der Funktionswerte verschieden.\n");
    exit(1);
  }
  max = 0;
  for (i=1; i <= n; i++) {
    q = fabs(y1->aktion->amIndex(y1, i, 1)
             - y2->aktion->amIndex(y2, i, 1));
    if (q>max) max = q;
  }
  fprintf(stdout, "Fehler = %lf\n", max);
 
  y1->aktion->freigeben(y1, NULL);
  y2->aktion->freigeben(y2, NULL);

  return 0;
}
