/* interplt.c
 * nuetzliche Funktionen fuer allgemeine interpolationen 
 * Hinweis:
 * Da wir noch keine gemeinsame Basisklasse fuer Matrizen und
 * Interpolationsaufgaben haben, sind hier einige Sachen doppelt.
 *
 * erstellt: Fre Nov 19 16:04:42 CET 1999
 * Autoren:  Mirko Belickin und Joachim Wagner
 */


#include "interplt.h"
#include "namen.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

/* externe Klassenobjekte einbinden */

extern interpolation newton_interpolation;
extern interpolation spline_interpolation;


/* Deklaration der Objekt-Methoden */

interpolation * interpolation_laden(FILE *datei, void *vater);


/* Globales Klassenobjekt */

interpolationFunktionen allg_interpolationFunktionen = {
    NULL,
    interpolation_laden
};

interpolation allg_interpolation = {        /* das Klassenobjekt zur Klasse ALLG_interpolation */
    &allg_interpolationFunktionen,
    NULL                      /* Klassenobjekt hat keine Daten */
};


/* ------------------------------------------------------------------------
 * Implementation
 */

/*
 * interpolation_laden
 * stellt den Typ der interpolation fest und veranlasst dann das
 * Laden der interpolation
 */

interpolation * interpolation_laden(FILE *datei, void *vater)
{
  char typ[80];
  typ[0] = '\0';                      /* leerer String */
  interpolation_kopfzeile(datei, typ);       /* interpolationtyp ermitteln */

  if (strcmp(typ, S_NEWTON_INTERPOLATION)==0) 
    return newton_interpolation.aktion->laden(datei, vater);

  if (strcmp(typ, S_SPLINE_INTERPOLATION)==0) 
    return spline_interpolation.aktion->laden(datei, vater);

  fprintf(stderr, "interpolation %s vom interpolationtyp %s kann nicht eingelesen "
                  "werden, da dieser Typ nicht im Klassenkluster "
                  "registriert.n", interpolation_kopfzeile(datei, typ), typ);
  exit(1);
}


/* interpolation_kopfzeile
 * liest die Kopfzeile eine interpolation ein, ueberprueft den Typ der interpolation
 * und gibt den Namen der interpolation als lokalen String zureuck, den sich
 * der Aufrufer zur weiteren Verwendung kopieren muss.
 * Wird ein leerer String als interpolationtyp uebergeben, so wird der Typ er-
 * mittelt und in interpolationtyp abgelegt. Ein nachfolgender Aufruf dieser
 * Funktion bearbeitet dann keine neue Zeile sondern liefert den Namen
 * der interpolation, deren Typ zuvor bestimmt wurde.

 * identisch zu matrix.c
 */

static int kopfzeile_halb_gelesen = 0;
static char eingabe[BUFSIZ];
static char unbenannt[] = "unbenannt";
static char *name;

char *interpolation_kopfzeile(FILE *datei, char *interpolationtyp) 
{
  char *p = NULL;

  if (!kopfzeile_halb_gelesen) {             /* Zeile noch nicht gelesen? */

    if (!fgets(eingabe,BUFSIZ,datei)) {          /* Zeile einlesen */
      fprintf(stderr,"Keine Daten fuer neue interpolation gefunden - EOF.\n");
      exit(1);
    }

    /* interpolationtyp feststellen */

    if (! (p = strchr(eingabe,':'))) {            /* Trennstelle suchen */
      fprintf(stderr,"Fehler in der Eingabe: "
                     "<interpolationtyp>:[interpolationname] erwartet.\n");
      exit(1);
    }
    *p = '\0';                                   /* String-Ende setzen */
  }

  if (*interpolationtyp) {                           /* interpolationtyp vorgegeben? */
    if (strcmp(eingabe, interpolationtyp)!=0) {      /* Ja -> ueberpruefen    */
      fprintf(stderr,"Dies ist keine %s interpolation!\n", interpolationtyp);
      exit(1);
    }
    if (kopfzeile_halb_gelesen) {
      kopfzeile_halb_gelesen = 0;
      return name;
    }
  }else{                                      /* Nein -> in den bereitge- */
    strcpy(interpolationtyp, eingabe);               /* stellten String kopieren */
    if (kopfzeile_halb_gelesen) {
      fprintf(stderr,"Fehler: Kopfzeile wurde bereits gelesen\n");
      exit(1);
    }
    kopfzeile_halb_gelesen = 1;
  }

  if (!p) {                                   /* Dieser Fehler kann nicht */
    fprintf(stderr,"Fehler in interpolation_kopfzeile\n");          /* auftreten */
    exit(1);
  } 

  /* Name der interpolation extrahieren */
  
  p++;                             /* hinter den Ex-':' schauen */
  while (*p && isspace(*p)) p++;   /* Anfang finden         */
  if (p[strlen(p)-1] == '\n')      /* ggf. abschliessendes NewLine tilgen */
    p[strlen(p)-1] = '\0';
  if (!*p) {                       /* Falls keinen gefunden */
    name = unbenannt;              /* default setzen        */
  }else{
    name = p;
  }
  return name;                     /* zurueckgeben */
}


/* interpolation_klasse
 * liefert Art der interpolation
 */

interpolation * interpolation_klasse(interpolation *m)
{
  if (m->aktion == newton_interpolation.aktion)
    return &newton_interpolation;
  fprintf(stderr, "Fehler: undefiniertes Objekt %s\n", m->aktion->name(m));
  exit(1);
}

