/* prueftab.c
 * Eingabe:  n + 1 Stuetzpunkte (Format siehe Beispieldateien)
 * Ausgabe:  Funktionswertetabelle zu einer der Beispielfunktionen
 * erstellt: Die Nov 23 16:43:51 CET 1999
 * Autoren : Mirko Belickin und Joachim Wagner
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "funktion.h"
#include "matrix.h"
#include "nmmatrix.h"
#include "interplt.h"
#include "newtnipl.h"

extern interpolation allg_interpolation;
extern matrix nm_matrix;

extern funktionsbeispiel funktion[];
extern const int funktion_anzahl;


/* Hauptprogramm 
 */

int main(int argc, char *argv[])
{
  FILE *out = stdout;
  FILE *temp;
  int hilfe = 0;        /* 1 = Hilfe fuer Benutzer anzeigen */   
  int fehler = 0;       /* 1 = Programm falsch aufgerufen */
  int schalter = 1;     /* 0 = es folgen keine Schalter ('--' gefunden) */
  int i, n, m;
  char tempname[140];
  matrix *y, *x;
  double a,b;           /* Grenzen des Intervalls */
  double stelle;
  nmmatrixschablone schablone;
  interpolation *polynom;
  
  while ((schalter) && (argc>1) && (*argv[1]=='-')) {/* Schalter bearbeiten */
    switch ((argv[1])[1]) {
      case 'h': hilfe = 1;
                break;
      case '-': schalter = 0;
                break;
      default:  fprintf(stderr,"Schalter %s unbekannt.\n", argv[1]);
                fehler = 1;
                break;
    }
    argv[1] = argv[0];
    argv++; 
    argc--;
  }
  
  if ((argc>4) || (argc<3)) fehler = 1;
  if ((hilfe) || (fehler)) {             /* Hilfetext anzeigen */
    fprintf(stderr,"\nBenutzung: "
                   "%s [Schalter] [--] n m [Ausgabedatei]\n\n",
                   argv[0]);
    fprintf(stderr,"Schalter: -h fuer diese Hilfe\n");
    fprintf(stderr," n+1 = Anzahl der Werte\n");
    fprintf(stderr," m = Nummer der Beispielfunktion\n");
    fprintf(stderr,"\nMirko Belickin und Joachim Wagner - 23. November 1999\n");
    exit(fehler);
  }
  if (argc>1) {
    n = atoi(argv[1]);
    if (n<1) {
      fprintf(stderr, "n muss mindestens 1 sein!\n");
      exit(1);
    }
  }
  if (argc>2) {
    m = atoi(argv[2]);
    if (m<1) {
      fprintf(stderr, "m muss mindestens 1 sein!\n");
      exit(1);
    }
    if (m>funktion_anzahl) {
      fprintf(stderr, "Funktion Nr. %i nicht definiert.\n", m);
      exit(1);
    }
  }
  if (argc>3) {
    if (!(temp=fopen(argv[3],"wb"))) {
      fprintf(stderr,"Oeffnen der Datei %s fehlgeschlagen.", argv[3]);
      exit(1);
    }
    out = temp;
  }
  
  schablone.n = n + 1;
  schablone.m = 1;
  schablone.init = 1;          /* mit 0 initialisieren */
  schablone.name = "X-Werte";
  x = nm_matrix.aktion->kreieren(&schablone, NULL);
  sprintf(tempname, "Y-Werte der Beispielfunktion %s", funktion[m-1].name);
  schablone.name = tempname;
  y = nm_matrix.aktion->kreieren(&schablone, NULL);
  a = funktion[m-1].a;
  b = funktion[m-1].b;
  if ((b - a) < EPSILON) {
      fprintf(stderr,"Intervall [%lf,%lf] numerisch zu klein oder "
                           "ungueltig.\n", a, b);
      exit(1);
  }
  for (i=0; i <= n; i++) {
    x->aktion->setzen(x, i+1, 1,
       stelle = (i * b + (n - i) * a) / n);
             /* umgeformt, so dass numerisch x(i) > b nicht auftritt */
    y->aktion->setzen(y, i+1, 1, funktion[m-1].wert(stelle));
  }
  
  x->aktion->speichern(x, out);
  y->aktion->speichern(y, out);
  x->aktion->freigeben(x, NULL);
  y->aktion->freigeben(y, NULL);

  if (out != stdout) fclose(out);               /* Dateien ggf. schliessen */
  return 0;
}
