/* splinipl.c
 * spline-Interpolation
 * erstellt: Fre Dez  3 15:20:27 CET 1999
 * Autoren:  Mirko Belickin und Joachim Wagner
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include "interplt.h"
#include "namen.h"
#include "matrix.h"
#include "nmmatrix.h"
#include "bandmtrx.h"
#include "a06_mtrx.h"
#include "splinipl.h"
#include "lgs.h"


#ifndef PI
  #define PI (4*atan(1))
#endif

/* externes Klassenobjekt einbinden */

extern matrix nm_matrix;
extern matrix band_matrix;
extern matrix a06__matrix;


/* Deklaration der Objekt-Methoden */

interpolation * splineinterpolation_laden(FILE *datei, void *vater);
int splineinterpolation_speichern(interpolation *m, FILE *datei);
interpolation * splineinterpolation_kreieren(void *schablone, void *vater);
int splineinterpolation_freigeben(interpolation *m, void *vater);
double splineinterpolation_funktionswert(interpolation *m, double x);
char *splineinterpolation_name(interpolation *m);
matrix * splineinterpolation_imVektor(interpolation *m, int vektorNr);
double splineinterpolation_linkeGrenze(interpolation *m);
double splineinterpolation_rechteGrenze(interpolation *m);
int splineinterpolation_koeffizienten(interpolation *m, void *vater); 

/* Globales Klassenobjekt */

interpolationFunktionen spline_interpolationFunktionen = {
    splineinterpolation_speichern,
    splineinterpolation_laden,
    splineinterpolation_freigeben,
    splineinterpolation_kreieren,
    splineinterpolation_funktionswert,
    splineinterpolation_name,
    interpolation_klasse,
    splineinterpolation_imVektor,
    splineinterpolation_linkeGrenze,
    splineinterpolation_rechteGrenze
};

interpolation spline_interpolation = {          /* das Klassenobjekt zur Klasse NM_interpolation */
    &spline_interpolationFunktionen,     
    NULL                      /* Klassenobjekt hat keine Daten */
};


/* private Methoden */

static splineinterpolationsdaten *get_splineinterpolationsdaten(interpolation *m);
static int lese_datum(FILE *datei, char *format, char * prefix, void *pdatum);
int splineinterpolation_pnichtperiodisch(interpolation *m, void *vater);
int splineinterpolation_periodisch(interpolation *m, void *vater);

/* ------------------------------------------------------------------------ 
 * spline-interpolation Implementation
 */ 


interpolation * splineinterpolation_laden(FILE *datei, void *vater)
{
  splineinterpolationsschablone schablone;
  interpolation *m;
  double a_ben, b_ben;

  schablone.name = interpolation_kopfzeile(datei, S_SPLINE_INTERPOLATION);

  lese_datum(datei, "%d", "stuetzstellenart =", 
                    &(schablone.stuetzstellenart));

  switch (schablone.stuetzstellenart) {
    case 0: schablone.x = nm_matrix.aktion->laden(datei, NULL);
            break;
    case 1: case 2:
            lese_datum(datei, "%lf", "a =", &(schablone.a));
            lese_datum(datei, "%lf", "b =", &(schablone.b));
            break;
  }

  lese_datum(datei, "%d", "splineart =", 
                    &(schablone.splineart));

  switch (schablone.splineart) {
    case 0: 
            lese_datum(datei, "%lf", "abl2links =", &(schablone.abl2links));
            lese_datum(datei, "%lf", "abl2rechts =", &(schablone.abl2rechts));
            break;
    case 1: 
    case 2: 
            break;
    case 3:
            lese_datum(datei, "%lf", "a_benutzer =", &a_ben);
            lese_datum(datei, "%lf", "b_benutzer =", &b_ben);
            break;
  }

  schablone.y = nm_matrix.aktion->laden(datei, NULL);

  /* interpolation erzeugen */

  if (schablone.splineart == 3) {
    splineinterpolationsdaten *pm;
    schablone.splineart = 2;
    m = spline_interpolation.aktion->kreieren(&schablone, vater);
    pm = get_splineinterpolationsdaten(m);
    schablone.splineart = 3;
    pm->a_ben = a_ben;
    pm->b_ben = b_ben;
  } else {
    m = spline_interpolation.aktion->kreieren(&schablone, vater);
  }

  /* Temporaere Daten wieder freigeben */

  if (schablone.stuetzstellenart==0)
    schablone.x->aktion->freigeben(schablone.x, NULL);
  schablone.y->aktion->freigeben(schablone.y, NULL);

  return m;
}

static int lese_datum(FILE *datei, char *format, char * prefix, void *pdatum)
{
  char eingabe[BUFSIZ];
  if (!fgets(eingabe,BUFSIZ,datei)) {
    fprintf(stderr,"Zeile %s nicht gefunden - EOF.\n", prefix);
    exit(1);
  }
  if (strncmp(eingabe, prefix,
               strlen( prefix))!=0) {
    fprintf(stderr,"%s... erwartet!\n", prefix);
    exit(1);
  }
  sscanf(&(eingabe[strlen(prefix)]), format, pdatum);
                               /* erstmal ohne Fehlerkontrolle */
  return 0;
}

int splineinterpolation_speichern(interpolation *m, FILE *datei)
{
  splineinterpolationsdaten *pm;
  int stuetzstellenart;
  pm = get_splineinterpolationsdaten(m);
  fprintf(datei,S_SPLINE_INTERPOLATION ": %s\n", pm->name);
  stuetzstellenart = pm->stuetzstellenart;
  stuetzstellenart &= (~INTERPOLATION_NUR_STELLEN_BERECHNEN);
  fprintf(datei,"stuetzstellenart = %i\n", stuetzstellenart);
  switch (stuetzstellenart) {
    case 0:
            pm->x->aktion->speichern(pm->x, datei);
            break;
    case 1: 
    case 2:
            fprintf(datei,"a = %lf\n", pm->a);
            fprintf(datei,"b = %lf\n", pm->b);
            break;
  }
  fprintf(datei,"splineart = %i\n", pm->splineart);
  switch (pm->splineart) {
    case 0:
            fprintf(datei,"abl2links = %lf\n", pm->abl2links);
            fprintf(datei,"abl2rechts = %lf\n", pm->abl2rechts);
            break;
    case 1: 
    case 2:
            break;
    case 3:
            fprintf(datei,"a = %lf\n", pm->a);
            fprintf(datei,"b = %lf\n", pm->b);
            fprintf(datei,"a_benutzer = %lf\n", pm->a_ben);
            fprintf(datei,"b_benutzer = %lf\n", pm->b_ben);
            break;
  }
  pm->y->aktion->speichern(pm->y, datei);
  return 0;
}

interpolation * splineinterpolation_kreieren(void *schablone, void *vater)
{
  splineinterpolationsschablone *s = (splineinterpolationsschablone *) schablone;
  interpolation *m;
  splineinterpolationsdaten *pm;
  char *name;
  int i, j, yfertig;
  nmmatrixschablone schablone2;
  double y1, yn1;
  int stuetzstellenart;
  matrix *y;

  if (!(m = (interpolation *)malloc(sizeof(interpolation)))) {
    fprintf(stderr,"Nicht genug Speicher fuer neue interpolation (Teil 1)!\n");
    exit(1);          
  }
  m->aktion = spline_interpolation.aktion;    /* moegliche Aktionen festlegen */
  if (!(m->daten = 
    (splineinterpolationsdaten *)malloc(sizeof(splineinterpolationsdaten)))) {
    fprintf(stderr,"Nicht genug Speicher fuer neue interpolation (Teil 2)!\n");
    exit(1);          
  }
  pm = get_splineinterpolationsdaten(m);

  /* Name der interpolation festlegen */

  name = (s && s->name)?s->name:"unbenannt";
  if (!(pm->name = (char *)malloc( sizeof(char) * (1+strlen(name))))) {
    fprintf(stderr, "Nicht genug Speicher fuer " S_SPLINE_INTERPOLATION 
                    " %s.\n", name);
    exit(1);
  }
  strcpy(pm->name, name);

  /* Groesse der interpolation festlegen */

  if (!s || !(s->y)) {
    fprintf(stderr, "Stuetzwerte fehlen fuer" S_SPLINE_INTERPOLATION " %s.\n",
                    name);
    exit(1);
  }
  pm->n = s->y->aktion->hoehe(s->y) - 1;
  if (pm->n < 0) {
    fprintf(stderr,"Es wird mindestens ein Stuetzwert benoetigt.\n");
    exit(1);
  }
  schablone2.n = pm->n + 1;
  schablone2.m = 1;
  schablone2.init = 0;
  schablone2.name = "Stuetzstellen";
  pm->x = nm_matrix.aktion->kreieren(&schablone2, m);
  schablone2.name = "Stuetzwerte";
  pm->y = nm_matrix.aktion->kreieren(&schablone2, m);
  schablone2.n = pm->n;
  schablone2.m = 4;
  schablone2.name = "Koeffizienten";
  pm->c = nm_matrix.aktion->kreieren(&schablone2, m);

  stuetzstellenart = s->stuetzstellenart;
  stuetzstellenart &= (~INTERPOLATION_NUR_STELLEN_BERECHNEN);

  switch(stuetzstellenart) {
    case 0:
          if (!(s->x)) {
            fprintf(stderr, "Stuetzstellen fehlen fuer" S_SPLINE_INTERPOLATION 
                            " %s.\n", name);
            exit(1);
          }
          if ((pm->n + 1) != (s->x->aktion->hoehe(s->x))) {
            fprintf(stderr, "Stuetzstellenanzahl/-werteanzahl inkompatibel " 
                            "fuer " S_SPLINE_INTERPOLATION " %s.\n", name);
            exit(1);
          }
          for (i=1; i <= (pm->n + 1); i++)
            pm->x->aktion->setzen(pm->x, i, 1,
              s->x->aktion->amIndex(s->x, i, 1));
          pm->stuetzstellenart = s->stuetzstellenart;
          break;

    case 1: /* aequidistante Stuetzstellen */
          if ((s->b - s->a) < EPSILON) {
            fprintf(stderr,"Intervall [%lf,%lf] numerisch zu klein oder "
                           "ungueltig.\n", s->a, s->b);
            exit(1);
          }
          for (i=0; i <= pm->n; i++)
            pm->x->aktion->setzen(pm->x, i+1, 1,
              (i * s->b + (pm->n - i) * s->a) / pm->n);   /* i-te Stuetz-   */
                                                          /* stelle, Formel */
                    /* umgeformt, so dass numerisch x(i) > b nicht auftritt */
          pm->stuetzstellenart = s->stuetzstellenart;
          pm->a = s->a;
          pm->b = s->b;
          break;
                
    case 2: /* Stuetzstellen nach Tschebyscheff-Polynom */
          if ((s->b - s->a) < EPSILON) {
            fprintf(stderr,"Intervall [%lf,%lf] numerisch zu klein oder "
                           "ungueltig.\n", s->a, s->b);
            exit(1);
          }
          for (i=0; i <= pm->n; i++)
            pm->x->aktion->setzen(pm->x, i+1, 1,
              (s->a + s->b)/2.0 + (s->a - s->b)*cos(i*PI/pm->n)/2.0);
          pm->stuetzstellenart = s->stuetzstellenart;
          pm->a = s->a;
          pm->b = s->b;
          break;
  }
  if (pm->n < 2) {  
    fprintf(stderr, "Zu wening Stuetzstellen fuer " 
                    S_SPLINE_INTERPOLATION 
                    " \'%s\'!\n",pm->name);
    exit(1);
  }
  yfertig = 0;
  switch(s->splineart) {
    case 0: 
            pm->abl2links = s->abl2links;
            pm->abl2rechts = s->abl2rechts;
            pm->splineart = 0;
            break;
    case 1:
            pm->abl2links = 0.0;    /* natuerliche Splines */
            pm->abl2rechts = 0.0;
            pm->splineart = 1;
            break;
    case 2:
            if ((pm->n == 2)        /* ggf. periodisch neue Punkte erzeugen */
              && !(pm->stuetzstellenart & INTERPOLATION_NUR_STELLEN_BERECHNEN))
            {
              int n;
              matrix *x, *c;
              double bma;
              fprintf(stderr, "Warnung: Nur %i Stuetzstellen fuer " 
                              S_SPLINE_INTERPOLATION 
                              " \'%s\', benutzte periodische Fortsetzung!\n",
                              pm->n + 1,  pm->name);
              n = 4; 
              schablone2.n = n + 1;
              schablone2.m = 1;
              schablone2.init = 0;
              schablone2.name = "Stuetzstellen";
              x = nm_matrix.aktion->kreieren(&schablone2, m);
              schablone2.name = "Stuetzwerte";
              y = nm_matrix.aktion->kreieren(&schablone2, m);
              schablone2.n = n;
              schablone2.m = 4;
              schablone2.name = "Koeffizienten";
              c = nm_matrix.aktion->kreieren(&schablone2, m);
              bma = pm->x->aktion->amIndex(pm->x, pm->n + 1, 1)
                    - pm->x->aktion->amIndex(pm->x, 1, 1);
              for (i=0; i<=1; i++) {
                for (j=1+i; j<=2+i; j++) {
                  x->aktion->setzen(x, i*2 + j, 1, 
                    i * bma + pm->x->aktion->amIndex(pm->x, j, 1));
                  y->aktion->setzen(y, i*2 + j, 1, 
                    s->y->aktion->amIndex(s->y, j, 1));
                }
              }
              x->aktion->setzen(x, 3, 1, 
                pm->x->aktion->amIndex(pm->x, 3, 1));
              y->aktion->setzen(y, 3, 1, 
                (s->y->aktion->amIndex(s->y, 1, 1)
                 + s->y->aktion->amIndex(s->y, 3, 1))/2.0);
              pm->x->aktion->freigeben(pm->x, m);
              pm->y->aktion->freigeben(pm->y, m);
              pm->c->aktion->freigeben(pm->c, m);
              if (pm->stuetzstellenart == 0) {
                pm->a = pm->x->aktion->amIndex(pm->x, 1, 1);
                pm->b = pm->x->aktion->amIndex(pm->x, pm->n+1, 1);
              }
              pm->x = x;
              pm->y = y;
              pm->c = c;
              pm->n = n;
              pm->a_ben = pm->a;
              pm->b_ben = pm->b;
              pm->b = pm->x->aktion->amIndex(pm->x, n+1, 1);
              yfertig = 1;
              fprintf(stderr,"Neue Stuetzpunkte:\n");
              pm->x->aktion->speichern(pm->x, stderr);
              pm->y->aktion->speichern(pm->y, stderr);
              pm->splineart = 3;
              y = pm->y;
            }else{
              pm->splineart = 2;
              y = s->y;
            }
            y1 = y->aktion->amIndex(y, 1, 1);
            yn1 = y->aktion->amIndex(y, pm->n+1, 1);
            if (fabs(y1-yn1) > EPSILON) {       
              fprintf(stderr, "Warnung: Anfangs- und Endstuetzwerte fuer " 
                              S_SPLINE_INTERPOLATION " \'%s\' auf %lf "
                              "gemittelt!\n", pm->name, (y1+yn1)/2.0);
              y->aktion->setzen(y, 1, 1, (y1+yn1)/2.0);
              y->aktion->setzen(y, pm->n+1, 1, (y1+yn1)/2.0);
            }
            break;
    default:
            fprintf(stderr, "Splineart %i nicht moeglich\n", 
                            s->splineart);
            exit(1); break;
  }
  if (!yfertig)
    for (i=1; i <= (pm->n + 1); i++)
      pm->y->aktion->setzen(pm->y, i, 1,
        s->y->aktion->amIndex(s->y, i, 1));
 
  /* Koeffizienten bestimmen */
  if (!(pm->stuetzstellenart & INTERPOLATION_NUR_STELLEN_BERECHNEN))
    splineinterpolation_koeffizienten(m, vater); 
  return m;
}

int splineinterpolation_freigeben(interpolation *m, void *vater) 
{ 
  splineinterpolationsdaten *pm;
  pm = get_splineinterpolationsdaten(m);
  free(pm->name);
  pm->x->aktion->freigeben(pm->x, m);
  pm->y->aktion->freigeben(pm->y, m);
  pm->c->aktion->freigeben(pm->c, m);
  free(pm);
  free(m);
  return 0;              /* immer Erfolg */
}

static splineinterpolationsdaten *get_splineinterpolationsdaten(interpolation *m) 
{
  if ((!(m->daten))) /* sicher ist sicher ... */
    exit(1);                        /* Fehlermeldung fehlt noch */
  return m->daten;                  /* Daten der interpolation zurueckgeben */
}

char *splineinterpolation_name(interpolation *m)
{
  splineinterpolationsdaten *pm;
  if (!m->daten) 
    return S_SPLINE_INTERPOLATION;
  pm = get_splineinterpolationsdaten(m);
  return pm->name; 
}

double splineinterpolation_funktionswert(interpolation *m, double x)
{
  splineinterpolationsdaten *pm;
  double p, xmxi;
  static double a,b;
  static int a_b_berechnet = 0;
  int links, rechts, mitte;
  int k;
  pm = get_splineinterpolationsdaten(m);

  /* Stelle x ins Intervall bringen */

  if (!a_b_berechnet) {
    a = splineinterpolation_linkeGrenze(m);
    b = splineinterpolation_rechteGrenze(m);
    a_b_berechnet = 1;
  }
  if ((pm->splineart==2) && ((x<a) || (x>b))) {
    /* fprintf(stderr,"Korrigieren x = %lf ", x); */
    while (x<a) x+=(b-a);    /* leider noetig */
    x = a + fmod(x-a, b-a);
    /* fprintf(stderr," zu %lf\n", x); */
  }
  
  /* Intervall ermitteln */
  links = 1;
  rechts = pm->n+1;
  while (rechts-links>1) {
    mitte = (links+rechts)>>1;
    if (pm->x->aktion->amIndex(pm->x, mitte, 1) < x) {
      links = mitte;
    }else{
      rechts = mitte;
    }
  }
  if (links <= pm->n) {
    /* Horner-Schema */
    xmxi = x - pm->x->aktion->amIndex(pm->x, links, 1);
    p = pm->c->aktion->amIndex(pm->c, links, 1);
    for (k= 2; k<=4; k++) {
      p *= xmxi;
      p += pm->c->aktion->amIndex(pm->c, links, k);         
    }
  }else{
    p = pm->y->aktion->amIndex(pm->y, rechts, 1);
  }
  return p;
}

matrix * splineinterpolation_imVektor(interpolation *m, int vektorNr)
{
  splineinterpolationsdaten *pm;
  pm = get_splineinterpolationsdaten(m);
  switch (vektorNr) {
    case 0: return pm->x;
    case 1: return pm->y;
    case 2: return pm->c;
    default:
    fprintf(stderr,"Ungueltiger Vektor fuer Interpolationsproblem %s: (%i)\n",
                   pm->name, vektorNr);
    exit(1); 
  }
}

double splineinterpolation_linkeGrenze(interpolation *m)
{
  splineinterpolationsdaten *pm;
  double min, h;
  int i;
  int n;
  pm = get_splineinterpolationsdaten(m);
  switch (pm->stuetzstellenart) {
    case 0:
        n = pm->x->aktion->hoehe(pm->x);
        min = pm->x->aktion->amIndex(pm->x, 1, 1);
        for (i=2; i<=n; i++) {
          h = pm->x->aktion->amIndex(pm->x, i, 1);
          if (h<min) min = h;
        }
        return min;
    case 1:
    case 2:
        if (pm->splineart==3)
          return pm->a_ben;
        else
          return pm->a;
    break;
    default:
        fprintf(stderr,"Ungueltige Stuetzstellenart fuer "
                       "Interpolationsproblem %s: (%i)\n",
                       pm->name, pm->stuetzstellenart);
        exit(1); 
  }
}

double splineinterpolation_rechteGrenze(interpolation *m)
{
  splineinterpolationsdaten *pm;
  double max, h;
  int i;
  int n;
  pm = get_splineinterpolationsdaten(m);
  switch (pm->stuetzstellenart) {
    case 0:
        n = pm->x->aktion->hoehe(pm->x);
        max = pm->x->aktion->amIndex(pm->x, 1, 1);
        for (i=2; i<=n; i++) {
          h = pm->x->aktion->amIndex(pm->x, i, 1);
          if (h>max) max = h;
        }
        return max;
    case 1:
    case 2:
        if (pm->splineart==3)
          return pm->b_ben;
        else
          return pm->b;
    break;
    default:
        fprintf(stderr,"Ungueltige Stuetzstellenart fuer "
                       "Interpolationsproblem %s: (%i)\n",
                       pm->name, pm->stuetzstellenart);
        exit(1); 
  }
}

int splineinterpolation_koeffizienten(interpolation *m, void *vater) 
{
  splineinterpolationsdaten *pm;
  pm = get_splineinterpolationsdaten(m);

  switch (pm->splineart) {
    case 1: /* weiterlaufen zum Fall 0 */
    case 0:
           splineinterpolation_pnichtperiodisch(m, vater);
           break;
    case 2:
    case 3:
           splineinterpolation_periodisch(m, vater);
           break;
    default:
           fprintf(stderr,"Splineart %i nicht implementiert.\n",
                          pm->splineart);
           exit(1);
  }
  return 0;   /* immer Erfolg (ausser wenn exit(1)) */
}

int splineinterpolation_pnichtperiodisch(interpolation *m, void *vater)
{
  splineinterpolationsdaten *pm;
  linearesSystem lgs;
  bandmatrixschablone schablone;
  nmmatrixschablone schablone2;
  matrix *h;                     /* Abstaende der Stuetzstellen */
  matrix *d1, *d2;               /* Nebendiagonalen */
  int i;
  double hilf, hilf2, hilf3;
  double yi, yi1, y2i, y2i1;

  pm = get_splineinterpolationsdaten(m);

  schablone.n = pm->n-1;     /* Tridiagonalmatrix erzeugen */
  schablone.m = 1;
  schablone.init = 0;
  schablone.name = "Hilfsmatrix kubische Splineinterpolation";
  lgs.a = band_matrix.aktion->kreieren(&schablone, m);
  
  schablone2.n = pm->n-1;    /* Vektor erzeugen */
  schablone2.m = 1;
  schablone2.init = 0;
  schablone2.name = "Hilfsvektor kubische Splineinterpolation (b)";
  lgs.b = nm_matrix.aktion->kreieren(&schablone2, m);

  schablone2.n = pm->n;      /* Vektor erzeugen */
  schablone2.m = 1;
  schablone2.init = 0;
  schablone2.name = "Hilfsvektor kubische Splineinterpolation (h)";
  h = nm_matrix.aktion->kreieren(&schablone2, m);

  for (i=1; i <= pm->n; i++) {  /* Intervallaengen berechnen */
    hilf = pm->x->aktion->amIndex(pm->x, i+1, 1) 
           - pm->x->aktion->amIndex(pm->x, i, 1) ;
    if (hilf < EPSILON) {
      fprintf(stderr, "Die Stuetzstellen fuer " S_SPLINE_INTERPOLATION 
                      " %s muessen aufsteigend sein!\n", pm->name);
      exit(1);
    }
    h->aktion->setzen(h, i, 1, hilf);
  }

  for (i=1; i <= pm->n-1; i++)         /* Hauptdiagonale belegen */
    lgs.a->aktion->setzen(lgs.a, i,i,
       2*(h->aktion->amIndex(h, i, 1) + 
          h->aktion->amIndex(h, i+1, 1)));

  d1 = lgs.a->aktion->imVektor(lgs.a, 1);
  d2 = lgs.a->aktion->imVektor(lgs.a, 2);

  for (i=1; i <= pm->n-2; i++) {       /* Nebendiagonale belegen */
    hilf = h->aktion->amIndex(h, i+1, 1);
    d1->aktion->setzen(d1, i, 1, hilf); 
    d2->aktion->setzen(d2, i, 1, hilf); 
  }
  
  hilf = pm->y->aktion->amIndex(pm->y, 2, 1)
         - pm->y->aktion->amIndex(pm->y, 1, 1);

  for (i=1; i <= pm->n-1; i++) {       /* b-Vektor bestimmen     */
    hilf2 = pm->y->aktion->amIndex(pm->y, i+2, 1)
            - pm->y->aktion->amIndex(pm->y, i+1, 1);
    hilf3 = 6 * (hilf/h->aktion->amIndex(h,i,1)
              - hilf2/h->aktion->amIndex(h,i+1,1));
    if (i==1) 
      hilf3 += pm->abl2links * h->aktion->amIndex(h, 1, 1);
    if (i==pm->n) 
      hilf3 += pm->abl2rechts * h->aktion->amIndex(h, pm->n, 1);

    lgs.b->aktion->setzen(lgs.b, i, 1, hilf3);
    hilf = hilf2;
  }
  
  if (!lrZerlegungUndDeterminanteUndLoesen(&lgs, diagonalStrategie)) {
    fprintf(stderr, "Fehler bei der Berechnung der Koeffizienten fuer " 
                    S_SPLINE_INTERPOLATION " %s!\n", pm->name);
    exit(1);
  } 

  /* jetzt sind die y" bekannt */

  schablone2.n = pm->n;      /* Koeffizientenmatrix erzeugen */
  schablone2.m = 4;
  schablone2.init = 0;
  schablone2.name = "Koeffizientenmatrix";
  pm->c = nm_matrix.aktion->kreieren(&schablone2, m);

  for (i=1; i<=pm->n; i++) {           /* Koeffizienten berechnen */
    yi  = pm->y->aktion->amIndex(pm->y, i,   1);
    yi1 = pm->y->aktion->amIndex(pm->y, i+1, 1);
    if (i==1) 
      y2i = pm->abl2links; 
    else
      y2i = lgs.x->aktion->amIndex(lgs.x, i-1, 1);
    if (i==pm->n) 
      y2i1 = pm->abl2rechts; 
    else
      y2i1 = lgs.x->aktion->amIndex(lgs.x, i, 1);

    hilf = h->aktion->amIndex(h, i, 1);

    pm->c->aktion->setzen(pm->c, i, 1, (y2i1 - y2i)/(6.0*hilf));
    pm->c->aktion->setzen(pm->c, i, 2, y2i / 2.0);
    pm->c->aktion->setzen(pm->c, i, 3, (yi1-yi)/hilf-hilf*(y2i1+2.0*y2i)/6.0);
    pm->c->aktion->setzen(pm->c, i, 4, yi);
  }

  /* noch zu erledigen: lgs freigeben */

  return 0;   /* immer Erfolg (ausser wenn exit(1)) */
 
}

int splineinterpolation_periodisch(interpolation *m, void *vater)
{
  splineinterpolationsdaten *pm;
  linearesSystem lgs;
  a06_matrixschablone schablone;
  nmmatrixschablone schablone2;
  matrix *h;                     /* Abstaende der Stuetzstellen */
  matrix *d1, *d2;               /* Nebendiagonalen */
  int i;
  double hilf, hilf2, hilf3;
  double yi, yi1, y2i, y2i1;

  pm = get_splineinterpolationsdaten(m);

  schablone.n = pm->n;       /* a06_matrix erzeugen */
  schablone.init = 0;
  schablone.name = "Hilfsmatrix kubische Splineinterpolation";
  lgs.a = a06__matrix.aktion->kreieren(&schablone, m);
  
  schablone2.n = pm->n;      /* Vektor erzeugen */
  schablone2.m = 1;
  schablone2.init = 0;
  schablone2.name = "Hilfsvektor kubische Splineinterpolation (b)";
  lgs.b = nm_matrix.aktion->kreieren(&schablone2, m);

  schablone2.n = pm->n;    /* Vektor erzeugen */
  schablone2.m = 1;
  schablone2.init = 0;
  schablone2.name = "Hilfsvektor kubische Splineinterpolation (h)";
  h = nm_matrix.aktion->kreieren(&schablone2, m);

  for (i=1; i <= pm->n; i++) {  /* Intervallaengen berechnen */
    hilf = pm->x->aktion->amIndex(pm->x, i+1, 1) 
           - pm->x->aktion->amIndex(pm->x, i, 1) ;
    if (hilf < EPSILON) {       
      fprintf(stderr, "Die Stuetzstellen fuer " S_SPLINE_INTERPOLATION 
                      " %s muessen aufsteigend sein!\n", pm->name);
      exit(1);
    }
    h->aktion->setzen(h, i, 1, hilf);
  }

  for (i=1; i <= pm->n; i++) {         /* Hauptdiagonale belegen */
    if (i==1)
      lgs.a->aktion->setzen(lgs.a, i,i,
         2*(h->aktion->amIndex(h, pm->n, 1) + 
            h->aktion->amIndex(h, i, 1)));
    else
      lgs.a->aktion->setzen(lgs.a, i,i,
         2*(h->aktion->amIndex(h, i-1, 1) + 
            h->aktion->amIndex(h, i, 1)));
  }

  d1 = lgs.a->aktion->imVektor(lgs.a, 1);
  d2 = lgs.a->aktion->imVektor(lgs.a, 2);

  for (i=1; i <= pm->n-1; i++) {       /* Nebendiagonale belegen */
    hilf = h->aktion->amIndex(h, i, 1);
    d1->aktion->setzen(d1, i, 1, hilf); 
  }
  
  hilf = h->aktion->amIndex(h, pm->n, 1);  /* Eckelement belegen */ 
  d2->aktion->setzen(d2, 1, 1, hilf); 

  hilf2 = pm->y->aktion->amIndex(pm->y, 1, 1)
          - pm->y->aktion->amIndex(pm->y, pm->n, 1);

  for (i=1; i <= pm->n; i++) {         /* b-Vektor bestimmen */
    
    hilf = pm->y->aktion->amIndex(pm->y, i+1, 1)
           - pm->y->aktion->amIndex(pm->y, i, 1);

    if (i==1)
      hilf3 = 6 * (-hilf/h->aktion->amIndex(h,i,1)
                + hilf2/h->aktion->amIndex(h,pm->n,1));
    else
      hilf3 = 6 * (-hilf/h->aktion->amIndex(h,i,1)
                + hilf2/h->aktion->amIndex(h,i-1,1));

    lgs.b->aktion->setzen(lgs.b, i, 1, hilf3);
    hilf2 = hilf;
  }
  
  if (!lrZerlegungUndDeterminanteUndLoesen(&lgs, diagonalStrategie)) {
    fprintf(stderr, "Fehler bei der Berechnung der Koeffizienten fuer " 
                    S_SPLINE_INTERPOLATION " %s!\n", pm->name);
    exit(1);
  } 

  /* jetzt sind die y" bekannt */

  schablone2.n = pm->n;    /* Koeffizientenmatrix erzeugen */
  schablone2.m = 4;
  schablone2.init = 0;
  schablone2.name = "Koeffizientenmatrix";
  pm->c = nm_matrix.aktion->kreieren(&schablone2, m);

  for (i=1; i<=pm->n; i++) {         /* Koeffizienten berechnen */
    yi  = pm->y->aktion->amIndex(pm->y, i,   1);
    yi1 = pm->y->aktion->amIndex(pm->y, i+1, 1);
    y2i = lgs.x->aktion->amIndex(lgs.x, i, 1);
    if (i==pm->n) 
      y2i1 = lgs.x->aktion->amIndex(lgs.x, 1, 1);
    else 
      y2i1 = lgs.x->aktion->amIndex(lgs.x, i+1, 1);

    hilf = h->aktion->amIndex(h, i, 1);

    pm->c->aktion->setzen(pm->c, i, 1, (y2i1 - y2i)/(6.0*hilf));
    pm->c->aktion->setzen(pm->c, i, 2, y2i / 2.0);
    pm->c->aktion->setzen(pm->c, i, 3, (yi1-yi)/hilf-hilf*(y2i1+2.0*y2i)/6.0);
    pm->c->aktion->setzen(pm->c, i, 4, yi);
  }

  /* noch zu erledigen: lgs freigeben */

  return 0;   /* immer Erfolg (ausser wenn exit(1)) */
 
}
