/* splinipl.c
 * spline-Interpolation
 * erstellt: Fre Dez  3 15:20:27 CET 1999
 * Autoren:  Mirko Belickin und Joachim Wagner
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include "interplt.h"
#include "namen.h"
#include "matrix.h"
#include "nmmatrix.h"
#include "bandmtrx.h"
#include "splinipl.h"
#include "lgs.h"


#ifndef PI
  #define PI (4*atan(1))
#endif

/* externes Klassenobjekt einbinden */

extern matrix nm_matrix;
extern matrix band_matrix;


/* Deklaration der Objekt-Methoden */

interpolation * splineinterpolation_laden(FILE *datei, void *vater);
int splineinterpolation_speichern(interpolation *m, FILE *datei);
interpolation * splineinterpolation_kreieren(void *schablone, void *vater);
int splineinterpolation_freigeben(interpolation *m, void *vater);
double splineinterpolation_funktionswert(interpolation *m, double x);
char *splineinterpolation_name(interpolation *m);
matrix * splineinterpolation_imVektor(interpolation *m, int vektorNr);
double splineinterpolation_linkeGrenze(interpolation *m);
double splineinterpolation_rechteGrenze(interpolation *m);
int splineinterpolation_koeffizienten(interpolation *m, void *vater); 

/* Globales Klassenobjekt */

interpolationFunktionen spline_interpolationFunktionen = {
    splineinterpolation_speichern,
    splineinterpolation_laden,
    splineinterpolation_freigeben,
    splineinterpolation_kreieren,
    splineinterpolation_funktionswert,
    splineinterpolation_name,
    interpolation_klasse,
    splineinterpolation_imVektor,
    splineinterpolation_linkeGrenze,
    splineinterpolation_rechteGrenze
};

interpolation spline_interpolation = {          /* das Klassenobjekt zur Klasse NM_interpolation */
    &spline_interpolationFunktionen,     
    NULL                      /* Klassenobjekt hat keine Daten */
};


/* private Methoden */

static splineinterpolationsdaten *get_splineinterpolationsdaten(interpolation *m);
static int lese_datum(FILE *datei, char *format, char * prefix, void *pdatum);
int splineinterpolation_pnichtperiodisch(interpolation *m, void *vater);

/* ------------------------------------------------------------------------ 
 * spline-interpolation Implementation
 */ 


interpolation * splineinterpolation_laden(FILE *datei, void *vater)
{
  splineinterpolationsschablone schablone;
  interpolation *m;

  schablone.name = interpolation_kopfzeile(datei, S_SPLINE_INTERPOLATION);

  lese_datum(datei, "%d", "stuetzstellenart =", 
                    &(schablone.stuetzstellenart));

  switch (schablone.stuetzstellenart) {
    case 0: schablone.x = nm_matrix.aktion->laden(datei, NULL);
            break;
    case 1: case 2:
            lese_datum(datei, "%lf", "a =", &(schablone.a));
            lese_datum(datei, "%lf", "b =", &(schablone.b));
            break;
  }

  lese_datum(datei, "%d", "splineart =", 
                    &(schablone.splineart));

  switch (schablone.splineart) {
    case 0: 
            lese_datum(datei, "%lf", "abl2links =", &(schablone.abl2links));
            lese_datum(datei, "%lf", "abl2rechts =", &(schablone.abl2rechts));
            break;
    case 1: 
            break;
  }

  schablone.y = nm_matrix.aktion->laden(datei, NULL);

  /* interpolation erzeugen */

  m = spline_interpolation.aktion->kreieren(&schablone, vater);

  /* Temporaere Daten wieder freigeben */
  if (schablone.stuetzstellenart==0)
    schablone.x->aktion->freigeben(schablone.x, NULL);
  schablone.y->aktion->freigeben(schablone.y, NULL);

  return m;
}

static int lese_datum(FILE *datei, char *format, char * prefix, void *pdatum)
{
  char eingabe[BUFSIZ];
  if (!fgets(eingabe,BUFSIZ,datei)) {
    fprintf(stderr,"Zeile %s nicht gefunden - EOF.\n", prefix);
    exit(1);
  }
  if (strncmp(eingabe, prefix,
               strlen( prefix))!=0) {
    fprintf(stderr,"%s... erwartet!\n", prefix);
    exit(1);
  }
  sscanf(&(eingabe[strlen(prefix)]), format, pdatum);
                               /* erstmal ohne Fehlerkontrolle */
  return 0;
}

int splineinterpolation_speichern(interpolation *m, FILE *datei)
{
  splineinterpolationsdaten *pm;
  pm = get_splineinterpolationsdaten(m);
  fprintf(datei,S_SPLINE_INTERPOLATION ": %s\n", pm->name);
  fprintf(datei,"stuetzstellenart = %i\n", pm->stuetzstellenart);
  switch (pm->stuetzstellenart) {
    case 0:
            pm->x->aktion->speichern(pm->x, datei);
            break;
    case 1: 
    case 2:
            fprintf(datei,"a = %lf\n", pm->a);
            fprintf(datei,"b = %lf\n", pm->b);
            break;
  }
  fprintf(datei,"splineart = %i\n", pm->splineart);
  switch (pm->splineart) {
    case 0:
            fprintf(datei,"abl2links = %lf\n", pm->abl2links);
            fprintf(datei,"abl2rechts = %lf\n", pm->abl2rechts);
            break;
    case 1: 
    case 2:
            break;
  }
  pm->y->aktion->speichern(pm->y, datei);
  return 0;
}

interpolation * splineinterpolation_kreieren(void *schablone, void *vater)
{
  splineinterpolationsschablone *s = (splineinterpolationsschablone *) schablone;
  interpolation *m;
  splineinterpolationsdaten *pm;
  char *name;
  int i;
  nmmatrixschablone schablone2;

  if (!(m = (interpolation *)malloc(sizeof(interpolation)))) {
    fprintf(stderr,"Nicht genug Speicher fuer neue interpolation (Teil 1)!\n");
    exit(1);          
  }
  m->aktion = spline_interpolation.aktion;    /* moegliche Aktionen festlegen */
  if (!(m->daten = 
    (splineinterpolationsdaten *)malloc(sizeof(splineinterpolationsdaten)))) {
    fprintf(stderr,"Nicht genug Speicher fuer neue interpolation (Teil 2)!\n");
    exit(1);          
  }
  pm = get_splineinterpolationsdaten(m);

  /* Name der interpolation festlegen */

  name = (s && s->name)?s->name:"unbenannt";
  if (!(pm->name = (char *)malloc( sizeof(char) * (1+strlen(name))))) {
    fprintf(stderr, "Nicht genug Speicher fuer " S_SPLINE_INTERPOLATION 
                    " %s.\n", name);
    exit(1);
  }
  strcpy(pm->name, name);

  /* Groesse der interpolation festlegen */

  if (!s || !(s->y)) {
    fprintf(stderr, "Stuetzwerte fehlen fuer" S_SPLINE_INTERPOLATION " %s.\n",
                    name);
    exit(1);
  }
  pm->n = s->y->aktion->hoehe(s->y) - 1;
  if (pm->n < 0) {
    fprintf(stderr,"Es wird mindestens ein Stuetzwert benoetigt.\n");
    exit(1);
  }
  schablone2.n = pm->n + 1;
  schablone2.m = 1;
  schablone2.init = 0;
  schablone2.name = "Stuetzstellen";
  pm->x = nm_matrix.aktion->kreieren(&schablone2, m);
  schablone2.name = "Stuetzwerte";
  pm->y = nm_matrix.aktion->kreieren(&schablone2, m);
  schablone2.n = pm->n;
  schablone2.m = 4;
  schablone2.name = "Koeffizienten";
  pm->c = nm_matrix.aktion->kreieren(&schablone2, m);

  switch(s->stuetzstellenart) {
    case 0:
          if (!(s->x)) {
            fprintf(stderr, "Stuetzstellen fehlen fuer" S_SPLINE_INTERPOLATION 
                            " %s.\n", name);
            exit(1);
          }
          if ((pm->n + 1) != (s->x->aktion->hoehe(s->x))) {
            fprintf(stderr, "Stuetzstellenanzahl/-werteanzahl inkompatibel " 
                            "fuer " S_SPLINE_INTERPOLATION " %s.\n", name);
            exit(1);
          }
          for (i=1; i <= (pm->n + 1); i++)
            pm->x->aktion->setzen(pm->x, i, 1,
              s->x->aktion->amIndex(s->x, i, 1));
          pm->stuetzstellenart = 0;
          break;

    case 1: /* aequidistante Stuetzstellen */
          if ((s->b - s->a) < EPSILON) {
            fprintf(stderr,"Intervall [%lf,%lf] numerisch zu klein oder "
                           "ungueltig.\n", s->a, s->b);
            exit(1);
          }
          for (i=0; i <= pm->n; i++)
            pm->x->aktion->setzen(pm->x, i+1, 1,
              (i * s->b + (pm->n - i) * s->a) / pm->n);   /* i-te Stuetz-   */
                                                          /* stelle, Formel */
                    /* umgeformt, so dass numerisch x(i) > b nicht auftritt */
          pm->stuetzstellenart = 1;
          pm->a = s->a;
          pm->b = s->b;
          break;
                
    case 2: /* Stuetzstellen nach Tschebyscheff-Polynom */
          if ((s->b - s->a) < EPSILON) {
            fprintf(stderr,"Intervall [%lf,%lf] numerisch zu klein oder "
                           "ungueltig.\n", s->a, s->b);
            exit(1);
          }
          for (i=0; i <= pm->n; i++)
            pm->x->aktion->setzen(pm->x, i+1, 1,
              (s->a + s->b)/2.0 + (s->a - s->b)*cos(i*PI/pm->n)/2.0);
          pm->stuetzstellenart = 2;
          pm->a = s->a;
          pm->b = s->b;
          break;
  }
  switch(s->splineart) {
    case 0: 
            pm->abl2links = s->abl2links;
            pm->abl2rechts = s->abl2rechts;
            pm->splineart = 0;
            break;
    case 1:
            pm->abl2links = 0.0;    /* natuerliche Splines */
            pm->abl2rechts = 0.0;
            pm->splineart = 1;
            break;
    default:
            fprintf(stderr, "Splineart %i nicht moeglich\n", 
                            s->splineart);
            exit(1); break;
  }
  for (i=1; i <= (pm->n + 1); i++)
    pm->y->aktion->setzen(pm->y, i, 1,
      s->y->aktion->amIndex(s->y, i, 1));
 
  /* Koeffizienten bestimmen */
  splineinterpolation_koeffizienten(m, vater); 
  return m;
}

int splineinterpolation_freigeben(interpolation *m, void *vater) 
{ 
  splineinterpolationsdaten *pm;
  pm = get_splineinterpolationsdaten(m);
  free(pm->name);
  pm->x->aktion->freigeben(pm->x, m);
  pm->y->aktion->freigeben(pm->y, m);
  pm->c->aktion->freigeben(pm->c, m);
  free(pm);
  free(m);
  return 0;              /* immer Erfolg */
}

static splineinterpolationsdaten *get_splineinterpolationsdaten(interpolation *m) 
{
  if ((!(m->daten))) /* sicher ist sicher ... */
    exit(1);                        /* Fehlermeldung fehlt noch */
  return m->daten;                  /* Daten der interpolation zurueckgeben */
}

char *splineinterpolation_name(interpolation *m)
{
  splineinterpolationsdaten *pm;
  if (!m->daten) 
    return S_SPLINE_INTERPOLATION;
  pm = get_splineinterpolationsdaten(m);
  return pm->name; 
}

double splineinterpolation_funktionswert(interpolation *m, double x)
{
  splineinterpolationsdaten *pm;
  double p, xmxi;
  static double a,b;
  static int a_b_berechnet = 0;
  int links, rechts, mitte;
  int k;
  double ai, bi, ci, di;
  pm = get_splineinterpolationsdaten(m);

  /* Stelle x ins Intervall bringen */

  if (!a_b_berechnet) {
    a = splineinterpolation_linkeGrenze(m);
    b = splineinterpolation_rechteGrenze(m);
    a_b_berechnet = 1;
  }
  if ((x<a) || (x>b)) {
    fprintf(stderr,"Korrigieren x = %lf ", x);
    x = a + fmod(x-a, b-a);
    fprintf(stderr," zu %lf\n", x); 
  }
  
  /* Intervall ermitteln */
  links = 1;
  rechts = pm->n+1;
  while (rechts-links>1) {
    mitte = (links+rechts)>>1;
    if (pm->x->aktion->amIndex(pm->x, mitte, 1) < x) {
      links = mitte;
    }else{
      rechts = mitte;
    }
  }
  if (links <= pm->n) {
    /* Horner-Schema */
    xmxi = x - pm->x->aktion->amIndex(pm->x, links, 1);
    p = pm->c->aktion->amIndex(pm->c, links, 1);
    for (k= 2; k<=4; k++) {
      p *= xmxi;
      p += pm->c->aktion->amIndex(pm->c, links, k);         
    }
    /*
    ai = pm->c->aktion->amIndex(pm->c, links, 1);
    bi = pm->c->aktion->amIndex(pm->c, links, 2);
    ci = pm->c->aktion->amIndex(pm->c, links, 3);
    di = pm->c->aktion->amIndex(pm->c, links, 4);
    p = ai * pow(xmxi,3) + bi * pow(xmxi,2) + ci * xmxi + di;
    */
  }else{
    p = pm->y->aktion->amIndex(pm->y, rechts, 1);
  }
  return p;
}

matrix * splineinterpolation_imVektor(interpolation *m, int vektorNr)
{
  splineinterpolationsdaten *pm;
  pm = get_splineinterpolationsdaten(m);
  switch (vektorNr) {
    case 0: return pm->x;
    case 1: return pm->y;
    case 2: return pm->c;
    default:
    fprintf(stderr,"Ungueltiger Vektor fuer Interpolationsproblem %s: (%i)\n",
                   pm->name, vektorNr);
    exit(1); 
  }
}

double splineinterpolation_linkeGrenze(interpolation *m)
{
  splineinterpolationsdaten *pm;
  double min, h;
  int i;
  int n;
  pm = get_splineinterpolationsdaten(m);
  switch (pm->stuetzstellenart) {
    case 0:
    n = pm->x->aktion->hoehe(pm->x);
    min = pm->x->aktion->amIndex(pm->x, 1, 1);
    for (i=2; i<=n; i++) {
      h = pm->x->aktion->amIndex(pm->x, i, 1);
      if (h<min) min = h;
    }
    return min;
    case 1:
    case 2:
    return pm->a;
    break;
    default:
    fprintf(stderr,"Ungueltige Stuetzstellenart fuer Interpolationsproblem %s: (%i)\n",
                   pm->name, pm->stuetzstellenart);
    exit(1); 
  }
}

double splineinterpolation_rechteGrenze(interpolation *m)
{
  splineinterpolationsdaten *pm;
  double max, h;
  int i;
  int n;
  pm = get_splineinterpolationsdaten(m);
  switch (pm->stuetzstellenart) {
    case 0:
    n = pm->x->aktion->hoehe(pm->x);
    max = pm->x->aktion->amIndex(pm->x, 1, 1);
    for (i=2; i<=n; i++) {
      h = pm->x->aktion->amIndex(pm->x, i, 1);
      if (h>max) max = h;
    }
    return max;
    case 1:
    case 2:
    return pm->b;
    break;
    default:
    fprintf(stderr,"Ungueltige Stuetzstellenart fuer Interpolationsproblem %s: (%i)\n",
                   pm->name, pm->stuetzstellenart);
    exit(1); 
  }
}

int splineinterpolation_koeffizienten(interpolation *m, void *vater) 
{
  splineinterpolationsdaten *pm;
  pm = get_splineinterpolationsdaten(m);

  if (pm->n < 2) {
    fprintf(stderr, "Mindestens 2 Stuetzstellen fuer " S_SPLINE_INTERPOLATION 
                    " %s noetig!\n", pm->name);
    exit(1);
  }

  switch (pm->splineart) {
    case 1: /* weiterlaufen zum Fall 0 */
    case 0:
           splineinterpolation_pnichtperiodisch(m, vater);
           break;
    case 2:
    default:
           fprintf(stderr,"Splineart %i nicht implementiert.\n",
                          pm->splineart);
           exit(1); break;
  }
}

int splineinterpolation_pnichtperiodisch(interpolation *m, void *vater)
{
  splineinterpolationsdaten *pm;
  linearesSystem lgs;
  bandmatrixschablone schablone;
  nmmatrixschablone schablone2;
  matrix *h;                     /* Abstaende der Stuetzstellen */
  matrix *d1, *d2;               /* Nebendiagonalen */
  int i;
  double hilf, hilf2, hilf3;
  double yi, yi1, y2i, y2i1;

  pm = get_splineinterpolationsdaten(m);

  schablone.n = pm->n-1;     /* Tridiagonalmatrix erzeugen */
  schablone.m = 1;
  schablone.init = 0;
  schablone.name = "Hilfsmatrix kubische Splineinterpolation";
  lgs.a = band_matrix.aktion->kreieren(&schablone, m);
  
  schablone2.n = pm->n-1;    /* Vektor erzeugen */
  schablone2.m = 1;
  schablone2.init = 0;
  schablone2.name = "Hilfsvektor kubische Splineinterpolation (b)";
  lgs.b = nm_matrix.aktion->kreieren(&schablone2, m);

  schablone2.n = pm->n;      /* Vektor erzeugen */
  schablone2.m = 1;
  schablone2.init = 0;
  schablone2.name = "Hilfsvektor kubische Splineinterpolation (h)";
  h = nm_matrix.aktion->kreieren(&schablone2, m);

  for (i=1; i <= pm->n; i++) {  /* Intervallaengen berechnen */
    hilf = pm->x->aktion->amIndex(pm->x, i+1, 1) 
           - pm->x->aktion->amIndex(pm->x, i, 1) ;
    if (hilf < EPSILON) {
      fprintf(stderr, "Die Stuetzstellen fuer " S_SPLINE_INTERPOLATION 
                      " %s muessen aufsteigend sein!\n", pm->name);
      exit(1);
    }
    h->aktion->setzen(h, i, 1, hilf);
  }

  for (i=1; i <= pm->n-1; i++)         /* Hauptdiagonale belegen */
    lgs.a->aktion->setzen(lgs.a, i,i,
       2*(h->aktion->amIndex(h, i, 1) + 
          h->aktion->amIndex(h, i+1, 1)));

  d1 = lgs.a->aktion->imVektor(lgs.a, 1);
  d2 = lgs.a->aktion->imVektor(lgs.a, 2);

  for (i=1; i <= pm->n-2; i++) {       /* Nebendiagonale belegen */
    hilf = h->aktion->amIndex(h, i+1, 1);
    d1->aktion->setzen(d1, i, 1, hilf); 
    d2->aktion->setzen(d2, i, 1, hilf); 
  }
  
  hilf = pm->y->aktion->amIndex(pm->y, 2, 1)
         - pm->y->aktion->amIndex(pm->y, 1, 1);

  for (i=1; i <= pm->n-1; i++) {       /* b-Vektor bestimmen     */
    hilf2 = pm->y->aktion->amIndex(pm->y, i+2, 1)
            - pm->y->aktion->amIndex(pm->y, i+1, 1);
    hilf3 = 6 * (hilf/h->aktion->amIndex(h,i,1)
              - hilf2/h->aktion->amIndex(h,i+1,1));
    if (i==1) 
      hilf3 += pm->abl2links * h->aktion->amIndex(h, 1, 1);
    if (i==pm->n) 
      hilf3 += pm->abl2rechts * h->aktion->amIndex(h, pm->n, 1);

    lgs.b->aktion->setzen(lgs.b, i, 1, hilf3);
    hilf = hilf2;
  }
  
  if (!lrZerlegungUndDeterminanteUndLoesen(&lgs, diagonalStrategie)) {
    fprintf(stderr, "Fehler bei der Berechnung der Koeffizienten fuer " 
                    S_SPLINE_INTERPOLATION " %s!\n", pm->name);
    exit(1);
  } 

  /* jetzt sind die y" bekannt */

  schablone2.n = pm->n;      /* Koeffizientenmatrix erzeugen */
  schablone2.m = 4;
  schablone2.init = 0;
  schablone2.name = "Koeffizientenmatrix";
  pm->c = nm_matrix.aktion->kreieren(&schablone2, m);

  for (i=1; i<=pm->n; i++) {           /* Koeffizienten berechnen */
    yi  = pm->y->aktion->amIndex(pm->y, i,   1);
    yi1 = pm->y->aktion->amIndex(pm->y, i+1, 1);
    if (i==1) 
      y2i = pm->abl2links; 
    else
      y2i = lgs.x->aktion->amIndex(lgs.x, i-1, 1);
    if (i==pm->n) 
      y2i1 = pm->abl2rechts; 
    else
      y2i1 = lgs.x->aktion->amIndex(lgs.x, i, 1);

    hilf = h->aktion->amIndex(h, i, 1);

    pm->c->aktion->setzen(pm->c, i, 1, (y2i1 - y2i)/(6.0*hilf));
    pm->c->aktion->setzen(pm->c, i, 2, y2i / 2.0);
    pm->c->aktion->setzen(pm->c, i, 3, (yi1-yi)/hilf-hilf*(y2i1+2.0*y2i)/6.0);
    pm->c->aktion->setzen(pm->c, i, 4, yi);
  }

  /* noch zu erledigen: lgs freigeben */

  return 0;   /* immer Erfolg (ausser wenn exit(1)) */
 
}
